# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

import re  # noqa: F401
import warnings
from typing import Optional

from truefoundry.ml._autogen.client.api_client import ApiClient
from truefoundry.ml._autogen.client.api_response import ApiResponse
from truefoundry.ml._autogen.client.exceptions import (  # noqa: F401
    ApiTypeError,
    ApiValueError,
)
from truefoundry.ml._autogen.client.models.body_get_search_runs_get import (
    BodyGetSearchRunsGet,
)
from truefoundry.ml._autogen.client.models.get_signed_urls_for_dataset_write_response_dto import (
    GetSignedURLsForDatasetWriteResponseDto,
)
from truefoundry.ml._autogen.client.models.run_response_dto import RunResponseDto
from truefoundry.ml._autogen.client.models.search_runs_response_dto import (
    SearchRunsResponseDto,
)
from truefoundry.pydantic_v1 import (
    StrictInt,
    StrictStr,
    conlist,
    validate_arguments,
)


class DeprecatedApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client

    @validate_arguments
    def get_run_by_name_get(
        self,
        run_name: StrictStr,
        experiment_id: Optional[StrictStr] = None,
        experiment_name: Optional[StrictStr] = None,
        **kwargs,
    ) -> RunResponseDto:  # noqa: E501
        """(Deprecated) Get Run By Name  # noqa: E501

        Get Run by run name  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_run_by_name_get(run_name, experiment_id, experiment_name, async_req=True)
        >>> result = thread.get()

        :param run_name: (required)
        :type run_name: str
        :param experiment_id:
        :type experiment_id: str
        :param experiment_name:
        :type experiment_name: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: RunResponseDto
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the get_run_by_name_get_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.get_run_by_name_get_with_http_info(
            run_name, experiment_id, experiment_name, **kwargs
        )  # noqa: E501

    @validate_arguments
    def get_run_by_name_get_with_http_info(
        self,
        run_name: StrictStr,
        experiment_id: Optional[StrictStr] = None,
        experiment_name: Optional[StrictStr] = None,
        **kwargs,
    ) -> ApiResponse:  # noqa: E501
        """(Deprecated) Get Run By Name  # noqa: E501

        Get Run by run name  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_run_by_name_get_with_http_info(run_name, experiment_id, experiment_name, async_req=True)
        >>> result = thread.get()

        :param run_name: (required)
        :type run_name: str
        :param experiment_id:
        :type experiment_id: str
        :param experiment_name:
        :type experiment_name: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(RunResponseDto, status_code(int), headers(HTTPHeaderDict))
        """

        warnings.warn(
            "GET /api/2.0mlflow/runs/get-by-name is deprecated.", DeprecationWarning
        )

        _params = locals()

        _all_params = ["run_name", "experiment_id", "experiment_name"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_run_by_name_get" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        if _params.get("run_name") is not None:  # noqa: E501
            _query_params.append(("run_name", _params["run_name"]))

        if _params.get("experiment_id") is not None:  # noqa: E501
            _query_params.append(("experiment_id", _params["experiment_id"]))

        if _params.get("experiment_name") is not None:  # noqa: E501
            _query_params.append(("experiment_name", _params["experiment_name"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "RunResponseDto",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0mlflow/runs/get-by-name",
            "GET",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def get_search_runs_get(
        self,
        filter: Optional[StrictStr] = None,
        page_token: Optional[StrictStr] = None,
        offset: Optional[StrictInt] = None,
        run_view_type: Optional[StrictStr] = None,
        max_results: Optional[StrictInt] = None,
        body_get_search_runs_get: Optional[BodyGetSearchRunsGet] = None,
        **kwargs,
    ) -> SearchRunsResponseDto:  # noqa: E501
        """(Deprecated) Get Search Runs  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_search_runs_get(filter, page_token, offset, run_view_type, max_results, body_get_search_runs_get, async_req=True)
        >>> result = thread.get()

        :param filter:
        :type filter: str
        :param page_token:
        :type page_token: str
        :param offset:
        :type offset: int
        :param run_view_type:
        :type run_view_type: str
        :param max_results:
        :type max_results: int
        :param body_get_search_runs_get:
        :type body_get_search_runs_get: BodyGetSearchRunsGet
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: SearchRunsResponseDto
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the get_search_runs_get_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.get_search_runs_get_with_http_info(
            filter,
            page_token,
            offset,
            run_view_type,
            max_results,
            body_get_search_runs_get,
            **kwargs,
        )  # noqa: E501

    @validate_arguments
    def get_search_runs_get_with_http_info(
        self,
        filter: Optional[StrictStr] = None,
        page_token: Optional[StrictStr] = None,
        offset: Optional[StrictInt] = None,
        run_view_type: Optional[StrictStr] = None,
        max_results: Optional[StrictInt] = None,
        body_get_search_runs_get: Optional[BodyGetSearchRunsGet] = None,
        **kwargs,
    ) -> ApiResponse:  # noqa: E501
        """(Deprecated) Get Search Runs  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_search_runs_get_with_http_info(filter, page_token, offset, run_view_type, max_results, body_get_search_runs_get, async_req=True)
        >>> result = thread.get()

        :param filter:
        :type filter: str
        :param page_token:
        :type page_token: str
        :param offset:
        :type offset: int
        :param run_view_type:
        :type run_view_type: str
        :param max_results:
        :type max_results: int
        :param body_get_search_runs_get:
        :type body_get_search_runs_get: BodyGetSearchRunsGet
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(SearchRunsResponseDto, status_code(int), headers(HTTPHeaderDict))
        """

        warnings.warn(
            "GET /api/2.0/preview/mlflow/runs/search is deprecated.", DeprecationWarning
        )

        _params = locals()

        _all_params = [
            "filter",
            "page_token",
            "offset",
            "run_view_type",
            "max_results",
            "body_get_search_runs_get",
        ]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_search_runs_get" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        if _params.get("filter") is not None:  # noqa: E501
            _query_params.append(("filter", _params["filter"]))

        if _params.get("page_token") is not None:  # noqa: E501
            _query_params.append(("page_token", _params["page_token"]))

        if _params.get("offset") is not None:  # noqa: E501
            _query_params.append(("offset", _params["offset"]))

        if _params.get("run_view_type") is not None:  # noqa: E501
            _query_params.append(("run_view_type", _params["run_view_type"]))

        if _params.get("max_results") is not None:  # noqa: E501
            _query_params.append(("max_results", _params["max_results"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["body_get_search_runs_get"] is not None:
            _body_params = _params["body_get_search_runs_get"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "SearchRunsResponseDto",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/preview/mlflow/runs/search",
            "GET",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def get_signed_urls_for_dataset_write_deprecated_get(
        self, dataset_fqn: StrictStr, paths: conlist(StrictStr), **kwargs
    ) -> GetSignedURLsForDatasetWriteResponseDto:  # noqa: E501
        """(Deprecated) Get Signed Urls For Dataset Write Deprecated  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_signed_urls_for_dataset_write_deprecated_get(dataset_fqn, paths, async_req=True)
        >>> result = thread.get()

        :param dataset_fqn: (required)
        :type dataset_fqn: str
        :param paths: (required)
        :type paths: List[str]
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: GetSignedURLsForDatasetWriteResponseDto
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the get_signed_urls_for_dataset_write_deprecated_get_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.get_signed_urls_for_dataset_write_deprecated_get_with_http_info(
            dataset_fqn, paths, **kwargs
        )  # noqa: E501

    @validate_arguments
    def get_signed_urls_for_dataset_write_deprecated_get_with_http_info(
        self, dataset_fqn: StrictStr, paths: conlist(StrictStr), **kwargs
    ) -> ApiResponse:  # noqa: E501
        """(Deprecated) Get Signed Urls For Dataset Write Deprecated  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_signed_urls_for_dataset_write_deprecated_get_with_http_info(dataset_fqn, paths, async_req=True)
        >>> result = thread.get()

        :param dataset_fqn: (required)
        :type dataset_fqn: str
        :param paths: (required)
        :type paths: List[str]
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(GetSignedURLsForDatasetWriteResponseDto, status_code(int), headers(HTTPHeaderDict))
        """

        warnings.warn(
            "GET /api/2.0/mlflow/mlfoundry-artifacts/datasets/get-signed-urls-for-write is deprecated.",
            DeprecationWarning,
        )

        _params = locals()

        _all_params = ["dataset_fqn", "paths"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_signed_urls_for_dataset_write_deprecated_get" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        if _params.get("dataset_fqn") is not None:  # noqa: E501
            _query_params.append(("dataset_fqn", _params["dataset_fqn"]))

        if _params.get("paths") is not None:  # noqa: E501
            _query_params.append(("paths", _params["paths"]))
            _collection_formats["paths"] = "multi"

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "GetSignedURLsForDatasetWriteResponseDto",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/mlfoundry-artifacts/datasets/get-signed-urls-for-write",
            "GET",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )



import os

DEBUG = os.getenv("AUTODEPLOY_DEBUG", "")

AUTODEPLOY_TFY_BASE_URL = os.getenv(
    "AUTODEPLOY_TFY_BASE_URL", "https://app.truefoundry.com"
).strip("/")
AUTODEPLOY_OPENAI_BASE_URL = os.environ.get("AUTODEPLOY_OPENAI_BASE_URL")
AUTODEPLOY_OPENAI_API_KEY = os.environ.get("AUTODEPLOY_OPENAI_API_KEY")
AUTODEPLOY_MODEL_NAME = os.environ.get(
    "AUTODEPLOY_MODEL_NAME", "auto-deploy-openai/gpt-4-turbo-2024-04-09"
)
ABOUT_AUTODEPLOY = """We'll use AI to build and deploy your project automatically.
Our AI Agent analyzes your codebase, checks for a Dockerfile, creates one if missing, builds a Docker image, fixes any issues, and runs the application to ensure we have built it correctly.
If you don't want to use our AI Agent to deploy automatically, create a [green]truefoundry.yaml[/] file in your project's root.
"""
# The maximum file size to read is set to 10KB.
# This limit is determined by the token limit of the LLM used, which is 128,000 tokens.
# Given that one token is approximately equivalent to 4 English characters,
# a 10KB file size limit (or ~10,000 characters |  ~2500 tokens) ensures that the file content,
# along with any additional context and instructions for the LLM, fits within the model's token limit.
MAX_FILE_SIZE_READ = 10 * 1024



from typing import Tuple

import rich_click as click

from truefoundry.cli.config import CliConfig
from truefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from truefoundry.cli.display_util import print_json
from truefoundry.cli.util import handle_exception_wrapper
from truefoundry.deploy.io.rich_output_callback import RichOutputCallBack
from truefoundry.deploy.lib.clients.servicefoundry_client import (
    ServiceFoundryServiceClient,
)
from truefoundry.deploy.lib.dao import application as application_lib
from truefoundry.deploy.lib.dao import delete as delete_lib
from truefoundry.deploy.lib.dao import workspace as workspace_lib
from truefoundry.deploy.lib.messages import (
    PROMPT_DELETED_APPLICATION,
    PROMPT_DELETED_WORKSPACE,
)
from truefoundry.deploy.lib.model.entity import File

# TODO (chiragjn): --json should disable all non json console prints


@click.group(name="delete", cls=GROUP_CLS, invoke_without_command=True)
@click.pass_context
@click.option(
    "-f",
    "--file",
    "files",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Path to yaml manifest file (You can pass multiple files at once by providing multiple -f options)",
    show_default=True,
    required=False,
    multiple=True,
)
@handle_exception_wrapper
def delete_command(ctx, files: Tuple[str, ...]):
    """
    Delete TrueFoundry resources
    """
    if ctx.invoked_subcommand is None:
        if not files:
            raise click.UsageError("Missing option '-f' / '--file'")
        sf_server_client = ServiceFoundryServiceClient()
        files_info = [File(path=file) for file in files]
        delete_results = delete_lib.delete_manifest_files(files_info, sf_server_client)

        if not all(delete_result.success for delete_result in delete_results):
            raise Exception("Failed to delete one or more resource manifests")


@click.command(name="workspace", cls=COMMAND_CLS, help="Delete a Workspace")
@click.option(
    "-w",
    "--workspace-fqn",
    "--workspace_fqn",
    type=click.STRING,
    help="FQN of the Workspace to delete",
    required=True,
)
@click.confirmation_option(prompt="Are you sure you want to delete this workspace?")
@handle_exception_wrapper
def delete_workspace(workspace_fqn):
    deleted_workspace = workspace_lib.delete_workspace(
        workspace_fqn=workspace_fqn,
    )
    output_hook = RichOutputCallBack()
    output_hook.print_line(PROMPT_DELETED_WORKSPACE.format(workspace_fqn))
    if CliConfig.json:
        print_json(data=deleted_workspace.dict())


@click.command(name="application", cls=COMMAND_CLS, help="Delete an Application")
@click.option(
    "--application-fqn",
    "--application_fqn",
    type=click.STRING,
    help="FQN of the Application to delete",
    required=True,
)
@click.confirmation_option(prompt="Are you sure you want to delete this application?")
@handle_exception_wrapper
def delete_application(application_fqn):
    response = application_lib.delete_application(
        application_fqn=application_fqn,
    )
    output_hook = RichOutputCallBack()
    output_hook.print_line(PROMPT_DELETED_APPLICATION.format(application_fqn))
    if CliConfig.json:
        print_json(data=response)


def get_delete_command():
    delete_command.add_command(delete_workspace)
    delete_command.add_command(delete_application)
    return delete_command



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional

from truefoundry.ml._autogen.client.models.subject_type import SubjectType
from truefoundry.pydantic_v1 import BaseModel, Field, StrictStr


class Subject(BaseModel):
    """
    Subject
    """

    subject_id: StrictStr = Field(default=..., alias="subjectId")
    subject_type: Optional[SubjectType] = Field(default=None, alias="subjectType")
    subject_slug: StrictStr = Field(default=..., alias="subjectSlug")
    subject_display_name: Optional[StrictStr] = Field(
        default=None, alias="subjectDisplayName"
    )
    __properties = ["subjectId", "subjectType", "subjectSlug", "subjectDisplayName"]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Subject:
        """Create an instance of Subject from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> Subject:
        """Create an instance of Subject from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return Subject.parse_obj(obj)

        _obj = Subject.parse_obj(
            {
                "subject_id": obj.get("subjectId"),
                "subject_type": obj.get("subjectType"),
                "subject_slug": obj.get("subjectSlug"),
                "subject_display_name": obj.get("subjectDisplayName"),
            }
        )
        return _obj



import logging
import sys

import rich_click as click

from truefoundry import logger
from truefoundry.cli.config import CliConfig
from truefoundry.cli.const import GROUP_CLS
from truefoundry.cli.util import setup_rich_click
from truefoundry.common.constants import TFY_DEBUG_ENV_KEY
from truefoundry.common.utils import is_internal_env_set
from truefoundry.deploy.cli.commands import (
    get_apply_command,
    get_build_command,
    get_delete_command,
    get_deploy_command,
    get_deploy_init_command,
    get_get_command,
    get_login_command,
    get_logout_command,
    get_patch_application_command,
    get_patch_command,
    get_terminate_command,
    get_trigger_command,
)
from truefoundry.ml.cli.cli import get_ml_cli
from truefoundry.version import __version__

click.rich_click.USE_RICH_MARKUP = True

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])  # noqa: C408


@click.group(
    cls=GROUP_CLS, context_settings=CONTEXT_SETTINGS, invoke_without_command=True
)
@click.option(
    "--json",
    is_flag=True,
    type=click.BOOL,
    help="Output entities in json format instead of formatted tables",
)
@click.option(
    "--debug",
    is_flag=True,
    type=click.BOOL,
    envvar=TFY_DEBUG_ENV_KEY,
    show_envvar=True,
    show_default=True,
    help="Set logging level to Debug. Can also be set using environment variable. E.g. TFY_DEBUG=1",
)
@click.version_option(
    version=__version__, message="TrueFoundry CLI: version %(version)s"
)
@click.pass_context
def truefoundry_cli(ctx, json, debug):
    """
    TrueFoundry provides an easy way to deploy your Services, Jobs and Models.
    \b

    To start, login to your TrueFoundry account with [code]tfy login[/]

    Then start deploying with [code]tfy deploy[/]

    Ask questions and troubleshoot your clusters with [code]tfy ask[/]

    And more: [link=https://docs.truefoundry.com/docs]https://docs.truefoundry.com/docs[/]

    """
    setup_rich_click()
    CliConfig.json = json
    CliConfig.debug = debug
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
    log_level = logging.INFO
    # no info logs while outputting json
    if json:
        log_level = logging.ERROR
    if debug:
        log_level = logging.DEBUG
    logger.add_cli_handler(level=log_level)


def create_truefoundry_cli() -> click.Group:
    """Generates CLI by combining all subcommands into a main CLI and returns in

    Returns:
        function: main CLI functions will all added sub-commands
    """
    cli = truefoundry_cli
    cli.add_command(get_login_command())
    cli.add_command(get_logout_command())
    cli.add_command(get_apply_command())
    cli.add_command(get_deploy_command())
    cli.add_command(get_deploy_init_command())
    cli.add_command(get_patch_application_command())
    cli.add_command(get_delete_command())
    cli.add_command(get_trigger_command())
    cli.add_command(get_terminate_command())
    cli.add_command(get_ml_cli())
    cli.add_command(get_get_command())

    if not (sys.platform.startswith("win32") or sys.platform.startswith("cygwin")):
        cli.add_command(get_patch_command())

    if is_internal_env_set():
        cli.add_command(get_build_command())
    return cli


def main():
    try:
        # We try and import readline to enable better prompt editing in the CLI
        import readline  # noqa: F401
    except ImportError:
        pass
    # Exit the interpreter by raising SystemExit(status).
    # If the status is omitted or None, it defaults to zero (i.e., success).
    # If the status is an integer, it will be used as the system exit status.
    # If it is another kind of object, it will be printed and the system exit status will be one (i.e., failure).
    try:
        cli = create_truefoundry_cli()
    except Exception as e:
        raise click.UsageError(message=str(e)) from e
    sys.exit(cli())


if __name__ == "__main__":
    main()



import logging
import math
import mmap
import os
from concurrent.futures import FIRST_EXCEPTION, Future, ThreadPoolExecutor, wait
from enum import Enum
from threading import Event
from typing import List, NamedTuple, Optional

import requests
from rich.progress import Progress
from tqdm.utils import CallbackIOWrapper

from truefoundry.common.constants import ENV_VARS
from truefoundry.common.exceptions import HttpRequestException
from truefoundry.common.request_utils import (
    augmented_raise_for_status,
    cloud_storage_http_request,
)
from truefoundry.pydantic_v1 import BaseModel

logger = logging.getLogger("truefoundry")


def truncate_path_for_progress(
    path: str, max_length: int, relpath: bool = False
) -> str:
    if relpath:
        path = os.path.relpath(path)
    if len(path) <= max_length:
        return path
    parts = path.split(os.sep)
    result = parts[-1]  # start with filename
    i = len(parts) - 2
    # keep prepending directories until adding more would exceed max_len - 3
    while i >= 0 and len(result) + len(parts[i]) + 1 <= max_length - 3:
        result = parts[i] + os.sep + result
        i -= 1
    return "..." + os.sep + result


class MultiPartUploadStorageProvider(str, Enum):
    S3_COMPATIBLE = "S3_COMPATIBLE"
    AZURE_BLOB = "AZURE_BLOB"


class SignedURL(BaseModel):
    signed_url: str
    path: Optional[str] = None


class MultiPartUpload(BaseModel):
    storage_provider: MultiPartUploadStorageProvider
    part_signed_urls: List[SignedURL]
    s3_compatible_upload_id: Optional[str] = None
    azure_blob_block_ids: Optional[List[str]] = None
    finalize_signed_url: SignedURL


class _FileMultiPartInfo(NamedTuple):
    num_parts: int
    part_size: int
    file_size: int


class _PartNumberEtag(NamedTuple):
    part_number: int
    etag: str


_MIN_BYTES_REQUIRED_FOR_MULTIPART = 100 * 1024 * 1024
# GCP/S3 Maximum number of parts per upload	10,000
# Maximum number of blocks in a block blob 50,000 blocks
# TODO: This number is artificially limited now. Later
# we will ask for parts signed URI in batches rather than in a single
# API Calls:
# Create Multipart Upload (Returns maximum number of parts, size limit of
#                            a single part, upload id for s3 etc )
#   Get me signed uris for first 500 parts
#     Upload 500 parts
#   Get me signed uris for the next 500 parts
#     Upload 500 parts
#   ...
# Finalize the Multipart upload using the finalize signed url returned
# by Create Multipart Upload or get a new one.
_MAX_NUM_PARTS_FOR_MULTIPART = 1000
# Azure Maximum size of a block in a block blob	4000 MiB
# GCP/S3 Maximum size of an individual part in a multipart upload 5 GiB
_MAX_PART_SIZE_BYTES_FOR_MULTIPART = 4 * 1024 * 1024 * 1000


class _CallbackIOWrapperForMultiPartUpload(CallbackIOWrapper):
    def __init__(self, callback, stream, method, length: int):
        self.wrapper_setattr("_length", length)
        super().__init__(callback, stream, method)

    def __len__(self):
        return self.wrapper_getattr("_length")


def _align_part_size_with_mmap_allocation_granularity(part_size: int) -> int:
    modulo = part_size % mmap.ALLOCATIONGRANULARITY
    if modulo == 0:
        return part_size

    part_size += mmap.ALLOCATIONGRANULARITY - modulo
    return part_size


# Can not be less than 5 * 1024 * 1024
_PART_SIZE_BYTES_FOR_MULTIPART = _align_part_size_with_mmap_allocation_granularity(
    10 * 1024 * 1024
)


def decide_file_parts(
    file_path: str,
    multipart_upload_allowed: bool = not ENV_VARS.TFY_ARTIFACTS_DISABLE_MULTIPART_UPLOAD,
    min_file_size_bytes_for_multipart: int = _MIN_BYTES_REQUIRED_FOR_MULTIPART,
) -> _FileMultiPartInfo:
    file_size = os.path.getsize(file_path)
    if not multipart_upload_allowed or file_size < min_file_size_bytes_for_multipart:
        return _FileMultiPartInfo(1, part_size=file_size, file_size=file_size)

    ideal_num_parts = math.ceil(file_size / _PART_SIZE_BYTES_FOR_MULTIPART)
    if ideal_num_parts <= _MAX_NUM_PARTS_FOR_MULTIPART:
        return _FileMultiPartInfo(
            ideal_num_parts,
            part_size=_PART_SIZE_BYTES_FOR_MULTIPART,
            file_size=file_size,
        )

    part_size_when_using_max_parts = math.ceil(file_size / _MAX_NUM_PARTS_FOR_MULTIPART)
    part_size_when_using_max_parts = _align_part_size_with_mmap_allocation_granularity(
        part_size_when_using_max_parts
    )
    if part_size_when_using_max_parts > _MAX_PART_SIZE_BYTES_FOR_MULTIPART:
        raise ValueError(
            f"file {file_path!r} is too big for upload. Multipart chunk"
            f" size {part_size_when_using_max_parts} is higher"
            f" than {_MAX_PART_SIZE_BYTES_FOR_MULTIPART}"
        )
    num_parts = math.ceil(file_size / part_size_when_using_max_parts)
    return _FileMultiPartInfo(
        num_parts, part_size=part_size_when_using_max_parts, file_size=file_size
    )


def _get_s3_compatible_completion_body(multi_parts: List[_PartNumberEtag]) -> str:
    body = "<CompleteMultipartUpload>\n"
    for part in multi_parts:
        body += "  <Part>\n"
        body += f"    <PartNumber>{part.part_number}</PartNumber>\n"
        body += f"    <ETag>{part.etag}</ETag>\n"
        body += "  </Part>\n"
    body += "</CompleteMultipartUpload>"
    return body


def _get_azure_blob_completion_body(block_ids: List[str]) -> str:
    body = "<BlockList>\n"
    for block_id in block_ids:
        body += f"<Uncommitted>{block_id}</Uncommitted> "
    body += "</BlockList>"
    return body


def _file_part_upload(
    url: str,
    file_path: str,
    seek: int,
    length: int,
    file_size: int,
    abort_event: Optional[Event] = None,
    method: str = "put",
    exception_class=HttpRequestException,
):
    def callback(*_, **__):
        if abort_event and abort_event.is_set():
            raise Exception("aborting upload")

    with open(file_path, "rb") as file:
        with mmap.mmap(
            file.fileno(),
            length=min(file_size - seek, length),
            offset=seek,
            access=mmap.ACCESS_READ,
        ) as mapped_file:
            wrapped_file = _CallbackIOWrapperForMultiPartUpload(
                callback, mapped_file, "read", len(mapped_file)
            )
            with cloud_storage_http_request(
                method=method,
                url=url,
                data=wrapped_file,
                exception_class=exception_class,
            ) as response:
                augmented_raise_for_status(response, exception_class=exception_class)
                return response


def s3_compatible_multipart_upload(  # noqa: C901
    multipart_upload: MultiPartUpload,
    local_file: str,
    multipart_info: _FileMultiPartInfo,
    executor: ThreadPoolExecutor,
    progress_bar: Optional[Progress] = None,
    abort_event: Optional[Event] = None,
    exception_class=HttpRequestException,
) -> None:
    abort_event = abort_event or Event()
    parts = []

    if progress_bar is not None:
        multi_part_upload_progress = progress_bar.add_task(
            f"[green]⬆ {truncate_path_for_progress(local_file, 64, relpath=True)}",
            start=True,
            visible=True,
        )

    def upload(part_number: int, seek: int) -> None:
        logger.debug(
            "Uploading part %d/%d of %s",
            part_number,
            multipart_info.num_parts,
            local_file,
        )
        response = _file_part_upload(
            url=multipart_upload.part_signed_urls[part_number].signed_url,
            file_path=local_file,
            seek=seek,
            length=multipart_info.part_size,
            file_size=multipart_info.file_size,
            abort_event=abort_event,
            exception_class=exception_class,
        )
        logger.debug(
            "Uploaded part %d/%d of %s",
            part_number,
            multipart_info.num_parts,
            local_file,
        )
        if progress_bar is not None:
            progress_bar.update(
                multi_part_upload_progress,
                advance=multipart_info.part_size,
                total=multipart_info.file_size,
            )
        etag = response.headers["ETag"]
        parts.append(_PartNumberEtag(etag=etag, part_number=part_number + 1))

    futures: List[Future] = []
    for part_number, seek in enumerate(
        range(0, multipart_info.file_size, multipart_info.part_size)
    ):
        future = executor.submit(upload, part_number=part_number, seek=seek)
        futures.append(future)

    done, not_done = wait(futures, return_when=FIRST_EXCEPTION)
    if len(not_done) > 0:
        abort_event.set()
    for future in not_done:
        future.cancel()
    for future in done:
        if future.exception() is not None:
            raise future.exception()

    logger.debug("Finalizing multipart upload of %s", local_file)
    parts = sorted(parts, key=lambda part: part.part_number)
    response = requests.post(
        multipart_upload.finalize_signed_url.signed_url,
        data=_get_s3_compatible_completion_body(parts),
        timeout=2 * 60,
    )
    response.raise_for_status()
    if progress_bar is not None:
        progress_bar.refresh()
    logger.debug("Multipart upload of %s completed", local_file)


def azure_multi_part_upload(  # noqa: C901
    multipart_upload: MultiPartUpload,
    local_file: str,
    multipart_info: _FileMultiPartInfo,
    executor: ThreadPoolExecutor,
    progress_bar: Optional[Progress] = None,
    abort_event: Optional[Event] = None,
    exception_class=HttpRequestException,
) -> None:
    abort_event = abort_event or Event()

    if progress_bar is not None:
        multi_part_upload_progress = progress_bar.add_task(
            f"[green]⬆ {truncate_path_for_progress(local_file, 64, relpath=True)}",
            start=True,
            visible=True,
        )

    def upload(part_number: int, seek: int):
        logger.debug(
            "Uploading part %d/%d of %s",
            part_number,
            multipart_info.num_parts,
            local_file,
        )
        _file_part_upload(
            url=multipart_upload.part_signed_urls[part_number].signed_url,
            file_path=local_file,
            seek=seek,
            length=multipart_info.part_size,
            file_size=multipart_info.file_size,
            abort_event=abort_event,
            exception_class=exception_class,
        )
        if progress_bar is not None:
            progress_bar.update(
                multi_part_upload_progress,
                advance=multipart_info.part_size,
                total=multipart_info.file_size,
            )
        logger.debug(
            "Uploaded part %d/%d of %s",
            part_number,
            multipart_info.num_parts,
            local_file,
        )

    futures: List[Future] = []
    for part_number, seek in enumerate(
        range(0, multipart_info.file_size, multipart_info.part_size)
    ):
        future = executor.submit(upload, part_number=part_number, seek=seek)
        futures.append(future)

    done, not_done = wait(futures, return_when=FIRST_EXCEPTION)
    if len(not_done) > 0:
        abort_event.set()
    for future in not_done:
        future.cancel()
    for future in done:
        if future.exception() is not None:
            raise future.exception()

    logger.debug("Finalizing multipart upload of %s", local_file)
    if multipart_upload.azure_blob_block_ids:
        response = requests.put(
            multipart_upload.finalize_signed_url.signed_url,
            data=_get_azure_blob_completion_body(
                block_ids=multipart_upload.azure_blob_block_ids
            ),
            timeout=2 * 60,
        )
        response.raise_for_status()
    if progress_bar is not None:
        progress_bar.refresh()
    logger.debug("Multipart upload of %s completed", local_file)



try:
    # pydantic >1.10.18
    from pydantic.v1 import *  # noqa: F403
    from pydantic.v1 import ConstrainedStr, utils  # noqa: F401
except ImportError:
    # pydantic <=1.10.17
    from pydantic import *  # noqa: F403
    from pydantic import ConstrainedStr, utils  # noqa: F401


class NonEmptyStr(ConstrainedStr):
    min_length: int = 1



from __future__ import annotations

import json
import os
import time
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, cast
from urllib.parse import urljoin

import requests
import socketio
from dateutil.tz import tzlocal
from rich.status import Status
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from truefoundry.common.constants import (
    SERVICEFOUNDRY_CLIENT_MAX_RETRIES,
    VERSION_PREFIX,
)
from truefoundry.common.request_utils import request_handling
from truefoundry.common.servicefoundry_client import (
    ServiceFoundryServiceClient as BaseServiceFoundryServiceClient,
)
from truefoundry.common.servicefoundry_client import (
    check_min_cli_version,
    session_with_retries,
)
from truefoundry.common.session import Session
from truefoundry.common.utils import get_user_agent
from truefoundry.deploy._autogen import models as autogen_models
from truefoundry.deploy.io.output_callback import OutputCallBack
from truefoundry.deploy.lib.model.entity import (
    Application,
    Cluster,
    CreateDockerRepositoryResponse,
    DependencyTreeResponse,
    Deployment,
    DockerRegistryCredentials,
    JobRun,
    LogBody,
    SocketEvent,
    TFYApplyResponse,
    TriggerJobResult,
    Workspace,
)
from truefoundry.deploy.lib.win32 import allow_interrupt
from truefoundry.deploy.v2.lib.models import (
    AppDeploymentStatusResponse,
    ApplicationFqnResponse,
    BuildResponse,
    DeploymentFqnResponse,
)
from truefoundry.logger import logger
from truefoundry.pydantic_v1 import parse_obj_as

DEPLOYMENT_LOGS_SUBSCRIBE_MESSAGE = "DEPLOYMENT_LOGS"
BUILD_LOGS_SUBSCRIBE_MESSAGE = "BUILD_LOGS"
MAX_RETRIES_WORKFLOW_TRIGGER = 3


def _upload_packaged_code(metadata: Dict[str, Any], package_file: str) -> None:
    file_size = os.stat(package_file).st_size
    with open(package_file, "rb") as file_to_upload:
        with tqdm(
            total=file_size,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            desc="Uploading package",
        ) as progress_bar:
            wrapped_file = CallbackIOWrapper(
                progress_bar.update, file_to_upload, "read"
            )
            headers = metadata.get("headers", {})
            http_response = requests.put(
                metadata["url"], data=wrapped_file, headers=headers
            )

            if http_response.status_code not in [204, 201, 200]:
                raise RuntimeError(f"Failed to upload code {http_response.content}")


class ServiceFoundryServiceClient(BaseServiceFoundryServiceClient):
    def __init__(
        self, init_session: bool = True, tfy_host: Optional[str] = None
    ) -> None:
        self._session: Optional[Session] = None
        if init_session:
            if tfy_host:
                logger.warning(f"Passed tfy_host {tfy_host!r} will be ignored")
            self._session = Session.new()
            tfy_host = self._session.tfy_host
        elif not tfy_host:
            raise Exception("Neither session, not tfy_host provided")
        super().__init__(tfy_host=tfy_host)

    def _get_headers(self) -> Dict[str, str]:
        headers = {"User-Agent": get_user_agent()}
        if not self._session:
            return headers
        return {
            **headers,
            "Authorization": f"Bearer {self._session.access_token}",
        }

    @check_min_cli_version
    def get_id_from_fqn(self, fqn_type: str, fqn: str) -> Dict[str, Any]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/fqn/{fqn_type}"
        response = session_with_retries().get(
            url, headers=self._get_headers(), params={"fqn": fqn}
        )
        return request_handling(response)

    @check_min_cli_version
    def list_clusters(self) -> List[Cluster]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/clusters"
        response = session_with_retries().get(url, headers=self._get_headers())
        response_data = cast(Dict[str, Any], request_handling(response))
        return parse_obj_as(List[Cluster], response_data["data"])

    @check_min_cli_version
    def list_workspaces(
        self,
        cluster_id: Optional[str] = None,
        workspace_name: Optional[str] = None,
        workspace_fqn: Optional[str] = None,
    ) -> List[Workspace]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspaces"
        params = {}
        if cluster_id:
            params["clusterId"] = cluster_id
        if workspace_name:
            params["workspaceName"] = workspace_name
        if workspace_fqn:
            params["workspaceFqn"] = workspace_fqn
        response = session_with_retries().get(
            url, params=params, headers=self._get_headers()
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return parse_obj_as(List[Workspace], response_data["data"])

    @check_min_cli_version
    def delete_workspace(self, workspace_id: str, force: bool = False) -> Workspace:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspaces/{workspace_id}"
        force_str = json.dumps(
            force
        )  # this dumb conversion is required because `params` just casts as str
        response = session_with_retries().delete(
            url, headers=self._get_headers(), params={"force": force_str}
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return Workspace.parse_obj(response_data["data"])

    @check_min_cli_version
    def get_workspace_by_fqn(self, workspace_fqn: str) -> List[Workspace]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspaces"
        response = session_with_retries().get(
            url,
            headers=self._get_headers(),
            params={"fqn": workspace_fqn},
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return parse_obj_as(List[Workspace], response_data["data"])

    @check_min_cli_version
    def get_cluster(self, cluster_id: str) -> Dict[str, Any]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/clusters/{cluster_id}"
        response = session_with_retries().get(url, headers=self._get_headers())
        response_data = cast(Dict[str, Any], request_handling(response))
        return response_data["data"]

    @check_min_cli_version
    def get_presigned_url(
        self, space_name: str, service_name: str, env: str
    ) -> Dict[str, Any]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/deployment/code-upload-url"
        response = session_with_retries().post(
            url,
            json={
                "workspaceFqn": space_name,
                "serviceName": service_name,
                "stage": env,
            },
            headers=self._get_headers(),
        )
        return cast(Dict[str, Any], request_handling(response))

    @check_min_cli_version
    def upload_code_package(
        self, workspace_fqn: str, component_name: str, package_local_path: str
    ) -> str:
        http_response = self.get_presigned_url(
            space_name=workspace_fqn, service_name=component_name, env="default"
        )
        _upload_packaged_code(metadata=http_response, package_file=package_local_path)
        return http_response["uri"]

    @check_min_cli_version
    def deploy_application(
        self,
        workspace_id: str,
        application: autogen_models.Workflow,
        force: bool = False,
        trigger_on_deploy: bool = False,
    ) -> Deployment:
        data = {
            "workspaceId": workspace_id,
            "name": application.name,
            # TODO (chiragjn): Should we also use exclude_unset=True here?
            "manifest": application.dict(exclude_none=True),
            "forceDeploy": force,
            "triggerOnDeploy": trigger_on_deploy,
        }
        logger.debug(json.dumps(data))
        url = f"{self._api_server_url}/{VERSION_PREFIX}/deployment"
        response = session_with_retries().post(
            url, json=data, headers=self._get_headers()
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return Deployment.parse_obj(response_data["deployment"])

    def _get_log_print_line(self, log_data: LogBody) -> str:
        timestamp = int(log_data.time) / 1e6

        time_obj = datetime.fromtimestamp(timestamp / 1000.0, tz=timezone.utc)
        time_obj.replace(tzinfo=timezone.utc)
        local_time = time_obj.astimezone(tzlocal())
        local_time_str = local_time.isoformat()
        return f"[{local_time_str}] {log_data.log.strip()}"

    def _tail_logs(
        self,
        tail_logs_url: str,
        query_dict: Dict[str, Any],
        # NOTE: Rather making this printer callback an argument,
        # we should have global printer callback
        # which will be initialized based on the running env (cli, lib, notebook)
        subscribe_message: str,
        socketio_path: str = "socket.io",
        callback: Optional[OutputCallBack] = None,
    ) -> socketio.Client:
        callback = callback or OutputCallBack()
        sio = socketio.Client(request_timeout=60, reconnection_attempts=10)
        callback.print_line("Waiting for the task to start...")
        next_log_start_timestamp = query_dict.get("startTs")

        @sio.on(subscribe_message)
        def logs(data: Any) -> None:
            try:
                event = SocketEvent.parse_obj(json.loads(data))
                if not isinstance(event.body, LogBody):
                    logger.debug(f"Skipped log for {data!r}")
                    return
                callback.print_line(self._get_log_print_line(event.body))
                nonlocal next_log_start_timestamp
                next_log_start_timestamp = event.body.time
            except Exception:
                logger.debug(f"Error while parsing log line, {data!r}")

        @sio.on("connect")
        def on_connect() -> None:
            # TODO: We should have have a timeout here. `emit` does
            #   not support timeout. Explore `sio.call`.
            query_dict["startTs"] = next_log_start_timestamp
            sio.emit(
                subscribe_message,
                json.dumps(query_dict),
            )

        def sio_disconnect_no_exception() -> None:
            try:
                sio.disconnect()
            except Exception:
                logger.exception("Error while disconnecting from socket connection")

        with allow_interrupt(sio_disconnect_no_exception):
            sio.connect(
                tail_logs_url,
                transports="websocket",
                headers=self._get_headers(),
                socketio_path=socketio_path,
                retry=True,
            )
            return sio

    @check_min_cli_version
    def get_deployment(self, application_id: str, deployment_id: str) -> Deployment:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/apps/{application_id}/deployments/{deployment_id}"
        response = session_with_retries().get(url, headers=self._get_headers())
        response_data = cast(Dict[str, Any], request_handling(response))
        return Deployment.parse_obj(response_data["data"])

    @check_min_cli_version
    def get_deployment_statuses(
        self,
        application_id: str,
        deployment_id: str,
        retry_count: int = SERVICEFOUNDRY_CLIENT_MAX_RETRIES,
    ) -> List[AppDeploymentStatusResponse]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/apps/{application_id}/deployments/{deployment_id}/statuses"
        response = session_with_retries(retries=retry_count).get(
            url, headers=self._get_headers()
        )
        response_data = request_handling(response)
        return parse_obj_as(List[AppDeploymentStatusResponse], response_data)

    @check_min_cli_version
    def get_deployment_build_response(
        self, application_id: str, deployment_id: str
    ) -> List[BuildResponse]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/apps/{application_id}/deployments/{deployment_id}/builds"
        response = session_with_retries().get(url, headers=self._get_headers())
        response_data = request_handling(response)
        return parse_obj_as(List[BuildResponse], response_data)

    def _get_deployment_logs(
        self,
        workspace_id: str,
        application_id: str,
        deployment_id: str,
        job_run_name: Optional[str] = None,
        start_ts_nano: Optional[int] = None,
        end_ts_nano: Optional[int] = None,
        limit: Optional[int] = None,
        num_logs_to_ignore: Optional[int] = None,
    ) -> List[LogBody]:
        get_logs_query = {"applicationId": application_id}
        if deployment_id:
            get_logs_query["deploymentId"] = deployment_id
        data: Dict[str, Any] = {"getLogsQuery": json.dumps(get_logs_query)}
        if start_ts_nano:
            data["startTs"] = str(start_ts_nano)
        if end_ts_nano:
            data["endTs"] = str(end_ts_nano)
        if limit:
            data["limit"] = str(limit)
        if num_logs_to_ignore:
            data["numLogsToIgnore"] = int(num_logs_to_ignore)
        if job_run_name:
            data["jobRunName"] = job_run_name

        url = f"{self._api_server_url}/{VERSION_PREFIX}/logs/{workspace_id}"
        response = session_with_retries().get(
            url=url, params=data, headers=self._get_headers()
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return parse_obj_as(List[LogBody], response_data["data"])

    @check_min_cli_version
    def tail_build_logs(
        self,
        build_response: BuildResponse,
        callback: Optional[OutputCallBack] = None,
    ) -> socketio.Client:
        callback = callback or OutputCallBack()
        tail_logs_obj = json.loads(build_response.tailLogsUrl)
        socket = self._tail_logs(
            tail_logs_url=urljoin(
                tail_logs_obj["uri"],
                f"/?type={BUILD_LOGS_SUBSCRIBE_MESSAGE}",
            ),
            socketio_path=tail_logs_obj["path"],
            query_dict={
                "pipelineRunName": build_response.name,
                "startTs": build_response.logsStartTs,
            },
            callback=callback,
            subscribe_message=BUILD_LOGS_SUBSCRIBE_MESSAGE,
        )
        return socket

    @check_min_cli_version
    def tail_logs_for_deployment(
        self,
        workspace_id: str,
        application_id: str,
        deployment_id: str,
        start_ts: int,
        limit: int,
        callback: Optional[OutputCallBack] = None,
        wait: bool = True,
    ) -> None:
        callback = callback or OutputCallBack()
        self._tail_logs(
            tail_logs_url=urljoin(
                self._api_server_url, f"/?type={DEPLOYMENT_LOGS_SUBSCRIBE_MESSAGE}"
            ),
            query_dict={
                "workspaceId": workspace_id,
                "startTs": str(int(start_ts * 1e6)),
                "limit": limit,
                "getLogsQuery": {
                    "applicationId": application_id,
                    "deploymentId": deployment_id,
                },
            },
            callback=callback,
            subscribe_message=DEPLOYMENT_LOGS_SUBSCRIBE_MESSAGE,
        )

    @check_min_cli_version
    def poll_logs_for_deployment(
        self,
        workspace_id: str,
        application_id: str,
        deployment_id: str,
        job_run_name: Optional[str],
        start_ts: int,
        limit: int,
        poll_interval_seconds: int,
        callback: Optional[OutputCallBack] = None,
    ) -> None:
        callback = callback or OutputCallBack()
        start_ts_nano = int(start_ts * 1e6)

        with Status(status="Polling for logs") as spinner:
            num_logs_to_ignore = 0

            while True:
                logs = self._get_deployment_logs(
                    workspace_id=workspace_id,
                    application_id=application_id,
                    deployment_id=deployment_id,
                    job_run_name=job_run_name,
                    start_ts_nano=start_ts_nano,
                    limit=limit,
                    num_logs_to_ignore=num_logs_to_ignore,
                )

                if not logs:
                    logger.warning("Did not receive any logs")
                    time.sleep(poll_interval_seconds)
                    continue

                for log in logs:
                    callback.print_line(self._get_log_print_line(log))

                last_log_time = logs[-1].time
                num_logs_to_ignore = 0
                for log in reversed(logs):
                    if log.time != last_log_time:
                        break
                    num_logs_to_ignore += 1

                start_ts_nano = int(last_log_time)
                spinner.update(status=f"Waiting for {poll_interval_seconds} secs.")
                time.sleep(poll_interval_seconds)

    @check_min_cli_version
    def fetch_deployment_logs(
        self,
        workspace_id: str,
        application_id: str,
        deployment_id: str,
        job_run_name: Optional[str],
        start_ts: Optional[int],
        end_ts: Optional[int],
        limit: Optional[int],
        callback: Optional[OutputCallBack] = None,
    ) -> None:
        callback = callback or OutputCallBack()
        logs = self._get_deployment_logs(
            workspace_id=workspace_id,
            application_id=application_id,
            deployment_id=deployment_id,
            job_run_name=job_run_name,
            start_ts_nano=int(start_ts * 1e6) if start_ts else None,
            end_ts_nano=int(end_ts * 1e6) if end_ts else None,
            limit=limit,
        )
        for log in logs:
            callback.print_line(self._get_log_print_line(log))

    @check_min_cli_version
    def fetch_build_logs(
        self,
        build_response: BuildResponse,
        callback: Optional[OutputCallBack] = None,
    ) -> None:
        callback = callback or OutputCallBack()
        url = build_response.getLogsUrl
        response = session_with_retries().get(url=url, headers=self._get_headers())
        logs_response = cast(Dict[str, Any], request_handling(response))
        for _log_body in logs_response["logs"]:
            try:
                log_body = LogBody.parse_obj(_log_body)
                callback.print_line(self._get_log_print_line(log_body))
            except Exception:
                logger.debug(f"Failed to parse log body {_log_body}")

    @check_min_cli_version
    def get_deployment_info_by_fqn(self, deployment_fqn: str) -> DeploymentFqnResponse:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/fqn/deployment"
        response = session_with_retries().get(
            url, headers=self._get_headers(), params={"fqn": deployment_fqn}
        )
        response_data = request_handling(response)
        return DeploymentFqnResponse.parse_obj(response_data)

    @check_min_cli_version
    def get_application_info_by_fqn(
        self, application_fqn: str
    ) -> ApplicationFqnResponse:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/fqn/app"
        response = session_with_retries().get(
            url, headers=self._get_headers(), params={"fqn": application_fqn}
        )
        response_data = request_handling(response)
        return ApplicationFqnResponse.parse_obj(response_data)

    @check_min_cli_version
    def delete_application(self, application_id: str) -> Dict[str, Any]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/apps/{application_id}"
        response = session_with_retries().delete(url, headers=self._get_headers())
        response_data = cast(Dict[str, Any], request_handling(response))
        # TODO: Add pydantic here.
        return response_data

    @check_min_cli_version
    def get_application_info(self, application_id: str) -> Application:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/apps/{application_id}"
        response = session_with_retries().get(url, headers=self._get_headers())
        response_data = cast(Dict[str, Any], request_handling(response))
        return Application.parse_obj(response_data["data"])

    def list_job_runs(
        self,
        application_id: str,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        search_prefix: Optional[str] = None,
    ) -> List[JobRun]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/jobs/{application_id}/runs"
        params = {}
        if limit:
            params["limit"] = limit
        if offset:
            params["offset"] = offset
        if search_prefix:
            params["searchPrefix"] = search_prefix
        response = session_with_retries().get(
            url, headers=self._get_headers(), params=params
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return parse_obj_as(List[JobRun], response_data["data"])

    def get_job_run(
        self,
        application_id: str,
        job_run_name: str,
    ) -> JobRun:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/jobs/{application_id}/runs/{job_run_name}"
        response = session_with_retries().get(url, headers=self._get_headers())
        response_data = cast(Dict[str, Any], request_handling(response))
        return JobRun.parse_obj(response_data["data"])

    def trigger_job(
        self,
        deployment_id: str,
        run_name_alias: Optional[str] = None,
        command: Optional[str] = None,
        params: Optional[Dict[str, str]] = None,
    ) -> TriggerJobResult:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/jobs/trigger"
        body = {
            "deploymentId": deployment_id,
            "input": {},
            "metadata": {},
        }
        if command:
            body["input"]["command"] = command
        if params:
            body["input"]["params"] = params
        if run_name_alias:
            body["metadata"]["job_run_name_alias"] = run_name_alias
        response = session_with_retries().post(
            url, json=body, headers=self._get_headers()
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return TriggerJobResult.parse_obj(response_data)

    def trigger_workflow(
        self, application_id: str, inputs: Dict[str, Any]
    ) -> Dict[str, Any]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workflow/{application_id}/executions"
        body = {"inputs": inputs}
        response = session_with_retries(retries=MAX_RETRIES_WORKFLOW_TRIGGER).post(
            url, json=body, headers=self._get_headers()
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return response_data

    @check_min_cli_version
    def get_docker_registry_creds(
        self, docker_registry_fqn: str, cluster_id: str
    ) -> DockerRegistryCredentials:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/docker-registry/creds"
        response = session_with_retries().get(
            url,
            headers=self._get_headers(),
            params={
                "fqn": docker_registry_fqn,
                "clusterId": cluster_id,
            },
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return DockerRegistryCredentials.parse_obj(response_data)

    @check_min_cli_version
    def create_repo_in_registry(
        self, docker_registry_fqn: str, workspace_fqn: str, application_name: str
    ) -> CreateDockerRepositoryResponse:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/docker-registry/create-repo"
        response = session_with_retries().post(
            url,
            headers=self._get_headers(),
            data={
                "fqn": docker_registry_fqn,
                "workspaceFqn": workspace_fqn,
                "applicationName": application_name,
            },
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return CreateDockerRepositoryResponse.parse_obj(response_data)

    @check_min_cli_version
    def list_applications(
        self,
        application_id: Optional[str] = None,
        workspace_id: Optional[str] = None,
        application_name: Optional[str] = None,
    ) -> List[Application]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/apps"
        params = {}
        if application_id:
            params["applicationId"] = application_id
        if workspace_id:
            params["workspaceId"] = workspace_id
        if application_name:
            params["applicationName"] = application_name
        response = session_with_retries().get(
            url, params=params, headers=self._get_headers()
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return parse_obj_as(List[Application], response_data["data"])

    @check_min_cli_version
    def list_versions(
        self,
        application_id: str,
        deployment_version: Optional[int] = None,
        deployment_id: Optional[str] = None,
    ) -> List[Deployment]:
        url = (
            f"{self._api_server_url}/{VERSION_PREFIX}/apps/{application_id}/deployments"
        )
        params = {}
        if deployment_version:
            params["version"] = deployment_version
        if deployment_id:
            params["deploymentId"] = deployment_id
        response = session_with_retries().get(
            url, params=params, headers=self._get_headers()
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return parse_obj_as(List[Deployment], response_data["data"])

    @check_min_cli_version
    def apply(
        self, manifest: Dict[str, Any], dry_run: bool = False
    ) -> TFYApplyResponse:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/apply"
        body = {"manifest": manifest, "dryRun": dry_run}
        response = session_with_retries().put(
            url, headers=self._get_headers(), json=body
        )
        response_data = cast(Dict[str, Any], request_handling(response) or {})
        return TFYApplyResponse.parse_obj(response_data)

    @check_min_cli_version
    def delete(self, manifest: Dict[str, Any]) -> Dict[str, Any]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/delete"
        body = {"manifest": manifest}
        response = session_with_retries().post(
            url, headers=self._get_headers(), json=body
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return response_data

    @check_min_cli_version
    def resolve_dependency_tree(
        self, manifests: List[Dict[str, Any]]
    ) -> DependencyTreeResponse:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/dependency-tree"
        body = {"manifests": manifests}
        response = session_with_retries().post(
            url, headers=self._get_headers(), json=body
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return DependencyTreeResponse.parse_obj(response_data)

    def terminate_job_run(
        self,
        deployment_id: str,
        job_run_name: str,
    ) -> Dict[str, Any]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/jobs/terminate"
        body = {
            "deploymentId": deployment_id,
            "jobRunName": job_run_name,
        }
        response = session_with_retries().post(
            url,
            # TODO (chiragjn): Check if this is supposed to be params or json
            params=body,
            json=body,
            headers=self._get_headers(),
        )
        response_data = cast(Dict[str, Any], request_handling(response))
        return response_data



try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata

try:
    __version__ = importlib_metadata.version("truefoundry")
except Exception:
    __version__ = "NA"



from __future__ import annotations

from typing import Any, ClassVar, Generator, Optional, Protocol, Type

from rich.console import Console

from truefoundry.pydantic_v1 import BaseModel


class Event(BaseModel):
    def render(self, _: Console) -> Optional[Any]: ...


class RequestEvent(Event): ...


class ResponseEvent(Event): ...


class Tool(Protocol):
    description: ClassVar[str]
    Request: ClassVar[Type[RequestEvent]]
    Response: ClassVar[Type[ResponseEvent]]

    def run(self, request: RequestEvent) -> Generator[Event, Any, ResponseEvent]: ...


class Message(Event):
    message: Any

    def render(self, console: Console):
        console.print(self.message)


class Rule(Event):
    message: Any

    def render(self, console: Console):
        console.rule(self.message)



import logging
import math
import os
from typing import TYPE_CHECKING, Any, Dict, Optional

import numpy as np

from truefoundry import ml

try:
    from transformers.integrations.integration_utils import rewrite_logs
    from transformers.trainer_callback import TrainerCallback
except ImportError as e:
    raise ImportError(
        "Importing this module requires `transformers` to be installed"
    ) from e

if TYPE_CHECKING:
    from transformers.trainer_callback import TrainerControl, TrainerState
    from transformers.training_args import TrainingArguments

    from truefoundry.ml import MlFoundryRun

logger = logging.getLogger(__name__)


class TrueFoundryMLCallback(TrainerCallback):
    def __init__(
        self,
        run: "MlFoundryRun",
        log_checkpoints: bool = True,
        checkpoint_artifact_name: Optional[str] = None,
        auto_end_run_on_train_end: bool = False,
        log_training_arguments: bool = True,
    ):
        """
        Args:
            run: The run entity to log metrics to.
            log_checkpoints: Whether to log checkpoints or not, defaults to True.
            checkpoint_artifact_name: The name of the artifact to log checkpoints to, required if log_checkpoints is True.
            auto_end_run_on_train_end: Whether to end the run automatically when training ends, defaults to False.
            log_training_arguments: Whether to log the training arguments or not, defaults to True.
            Usage:
                from transformers import Trainer
                from truefoundry.ml.integrations.huggingface.trainer_callback import TrueFoundryMLCallback
                from truefoundry.ml import get_client

                client = get_client()
                run = client.create_run(ml_repo="my-ml-repo", run_name="my-run", auto_end=False)

                callback = TrueFoundryMLCallback(
                    run=run,
                    log_checkpoints=True,
                    checkpoint_artifact_name="my-checkpoint",
                    auto_end_run_on_train_end=True,
                )

                trainer = Trainer(
                    ...,
                    callbacks=[callback]
                )
        """
        self._run = run
        self._log_checkpoints = log_checkpoints
        if self._log_checkpoints and not checkpoint_artifact_name:
            raise ValueError(
                "`checkpoint_artifact_name` is required when `log_checkpoints` is True"
            )
        self._checkpoint_artifact_name = checkpoint_artifact_name
        self._auto_end_run_on_train_end = auto_end_run_on_train_end
        self._log_training_arguments = log_training_arguments

    @classmethod
    def with_managed_run(
        cls,
        ml_repo: str,
        run_name: Optional[str] = None,
        log_checkpoints: bool = True,
        checkpoint_artifact_name: Optional[str] = None,
        auto_end_run_on_train_end: bool = True,
        log_training_arguments: bool = True,
    ) -> "TrueFoundryMLCallback":
        """
        Args:
            ml_repo: The name of the ML Repository to log metrics and data to.
            run_name: The name of the run, if not provided, a random name will be generated.
            log_checkpoints: Whether to log checkpoints or not, defaults to True.
            checkpoint_artifact_name: The name of the artifact to log checkpoints to, required if log_checkpoints is True.
            auto_end_run_on_train_end: Whether to end the run automatically when training ends, defaults to True.
            log_training_arguments: Whether to log the training arguments or not, defaults to True.
        Usage:
            from transformers import Trainer
            from truefoundry.ml.integrations.huggingface.trainer_callback import TrueFoundryMLCallback

            callback = TrueFoundryMLCallback.with_managed_run(
                ml_repo="my-ml-repo",
                run_name="my-run",
                log_checkpoints=True,
                checkpoint_artifact_name="my-checkpoint",
                auto_end_run_on_train_end=True,
            )
            trainer = Trainer(
                ...,
                callbacks=[callback]
            )
        """
        run = ml.get_client().create_run(
            ml_repo=ml_repo, run_name=run_name, auto_end=False
        )
        return cls(
            run=run,
            log_checkpoints=log_checkpoints,
            checkpoint_artifact_name=checkpoint_artifact_name,
            auto_end_run_on_train_end=auto_end_run_on_train_end,
            log_training_arguments=log_training_arguments,
        )

    def _drop_non_finite_values(self, dct: Dict[str, Any]) -> Dict[str, Any]:
        sanitized = {}
        for k, v in dct.items():
            if isinstance(v, (int, float, np.integer, np.floating)) and math.isfinite(
                v
            ):
                sanitized[k] = v
            else:
                logger.warning(
                    f'Trainer is attempting to log a value of "{v}" of'
                    f' type {type(v)} for key "{k}" as a metric.'
                    " Mlfoundry's log_metric() only accepts finite float and"
                    " int types so we dropped this attribute."
                )
        return sanitized

    @property
    def run(self) -> "MlFoundryRun":
        return self._run

    # noinspection PyMethodOverriding
    def on_log(
        self,
        args: "TrainingArguments",
        state: "TrainerState",
        control: "TrainerControl",
        logs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        logs = logs or {}
        if not state.is_world_process_zero:
            return

        metrics = self._drop_non_finite_values(logs)
        self._run.log_metrics(rewrite_logs(metrics), step=state.global_step)

    def on_save(
        self,
        args: "TrainingArguments",
        state: "TrainerState",
        control: "TrainerControl",
        **kwargs,
    ):
        if not state.is_world_process_zero:
            return

        if not self._log_checkpoints:
            return

        if not self._checkpoint_artifact_name:
            return

        ckpt_dir = f"checkpoint-{state.global_step}"
        artifact_path = os.path.join(args.output_dir, ckpt_dir)
        description = None
        _job_name = os.getenv("TFY_INTERNAL_COMPONENT_NAME")
        _job_run_name = os.getenv("TFY_INTERNAL_JOB_RUN_NAME")
        if _job_name:
            description = f"Checkpoint from job={_job_name} run={_job_run_name}"
        logger.info(f"Uploading checkpoint {ckpt_dir} ...")
        metadata = {}
        for log in state.log_history:
            if isinstance(log, dict) and log.get("step") == state.global_step:
                metadata = log.copy()
        metadata = self._drop_non_finite_values(metadata)
        self._run.log_artifact(
            name=self._checkpoint_artifact_name,
            artifact_paths=[(artifact_path, None)],
            metadata=metadata,
            step=state.global_step,
            description=description,
        )

    def on_train_begin(
        self,
        args: "TrainingArguments",
        state: "TrainerState",
        control: "TrainerControl",
        **kwargs,
    ):
        if self._log_training_arguments:
            training_arguments = args.to_sanitized_dict()
            valid_types = (bool, int, float, str)
            training_arguments = {
                k: v for k, v in training_arguments.items() if type(v) in valid_types
            }
            self._run.log_params(training_arguments)

    def on_train_end(
        self,
        args: "TrainingArguments",
        state: "TrainerState",
        control: "TrainerControl",
        **kwargs,
    ):
        """
        Event called at the end of training.
        """
        if self._auto_end_run_on_train_end:
            self._run.end()



# deprecated
from __future__ import annotations

import os
from typing import Optional

from rich.console import Console
from rich.prompt import Confirm

from truefoundry.autodeploy.tools.base import Tool
from truefoundry.autodeploy.utils.diff import Diff
from truefoundry.pydantic_v1 import BaseModel, Field

INTERACTIVE_SESSION = True


class WriteFile(Tool):
    description = """
Write contents to a file.
Read first before write.
"""

    class Request(BaseModel):
        path: str = Field(
            ...,
            pattern=r"^[a-zA-Z0-9\.]{1}.*$",
            description="File path to write.",
        )
        content: str = Field(..., description="Content of the file.")
        justification: str = Field(
            ...,
            description="Justification of why the new content is required. Ensure the justification, justifies the content.",
        )

    class Response(BaseModel):
        cancellation_reason: Optional[str] = Field(
            None, description="Operation cancelled by user"
        )

        error: Optional[str] = Field(
            None,
            description="Error while opening a file.",
        )

    def __init__(self, project_root_path: str):
        self.project_root_path = project_root_path

    def _user_interaction(
        self, request: WriteFile.Request, console: Console
    ) -> Optional[WriteFile.Response]:
        console.log("You are about to write or edit a file.")

        console.print(f"Displaying changes in {request.path}")
        prev_content = ""

        if os.path.exists(os.path.join(self.project_root_path, request.path)):
            with open(
                os.path.join(self.project_root_path, request.path),
                "r",
                encoding="utf8",
            ) as f:
                prev_content = f.read()

        console.print(
            Diff(
                lhs=prev_content,
                rhs=request.content,
                width=os.get_terminal_size().columns,
            )
        )
        response = Confirm.ask(
            f"Writing file at {request.path}?",
        )
        if not response:
            description = console.input(
                "You chose to cancel. Can you provide a reason why? [green]>> "
            )
            return WriteFile.Response(
                cancellation_reason=description, error="Operation cancelled by user."
            )
        console.log(f"Writing file to {request.path}")

    def run(
        self,
        request: WriteFile.Request,
        console: Console,
    ) -> WriteFile.Response:
        if INTERACTIVE_SESSION:
            interaction_response = self._user_interaction(request, console)
            if isinstance(interaction_response, WriteFile.Response):
                return interaction_response
        try:
            with open(
                os.path.join(self.project_root_path, request.path),
                "w",
                encoding="utf8",
            ) as f:
                f.write(request.content)
                return WriteFile.Response()
        except FileNotFoundError as ex:
            return WriteFile.Response(error=str(ex))
